/*
 * DANG_BEGIN_MODULE
 *
 * This module contains the code to handle the signals telling DOSEMU
 * to switch in and out of the current console.
 *
 * DANG_END_MODULE
 *
 * DANG_BEGIN_CHANGELOG
 *
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 *
 * DANG_END_CHANGELOG
 */


#include "dosemu.h"
#include "emu_mem.h"
#include "emu_termio.h"
#include "emu_mouse.h"
#include "emu_bios.h"
#include "emu_port.h"
#include "emu_video.h"
#include "emu_detach.h"
#include "vc.h"
#include "arch/signal.h"

#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/kd.h>
#include <linux/vt.h>

static void allow_switch (void);

void scr_state_init(void)
{
  scr_state.vt_allow = 0;
  scr_state.vt_requested = 0;
  scr_state.mapped = 0;
  scr_state.pageno = 0;
  scr_state.virt_address = PAGE_ADDR(0);
  scr_state.current = 1;
}

static void parent_close_mouse(void)
{
  if (mice->intdrv) {
    remove_from_io_select(mice->fd, mice->add_to_io_select);
    DOS_SYSCALL(close (mice->fd));
  }
  else
    child_close_mouse();
}

static void parent_open_mouse(void)
{
  if (mice->intdrv) {
    mice->fd = DOS_SYSCALL (open (mice->dev, O_RDWR | O_NONBLOCK));
    if (mice->fd > 0)
      add_to_io_select(mice->fd, mice->add_to_io_select);
  }
  else
    child_open_mouse();
}

static void SIGACQUIRE_call(void)
{
  vc_attach();
  parent_open_mouse();
}

static inline void forbid_switch(void)
{
  v_printf("VID: forbid_switch\n");
  scr_state.vt_allow = 0;
}

static void acquire_vt(int sig, struct sigcontext_struct context)
{
  v_printf("VID: Acquiring VC\n");
  forbid_switch();
  if (ioctl(kbd_fd, VT_RELDISP, VT_ACKACQ))	/* switch acknowledged */
    v_printf("VID: VT_RELDISP failed (or was queued)!\n");
  allow_switch();
  signal_queue(SIGACQUIRE_call);
  scr_state.current = 1;
}

static void SIGRELEASE_call (void)
{
  if (scr_state.current == 1) {
    v_printf("VID: Releasing VC\n");
    if (!scr_state.vt_allow) {
      v_printf("VID: Disallowed vt switch!\n");
      return;
    }
    parent_close_mouse ();
    vc_detach();
  }

  scr_state.current = 0;	/* our console is no longer current */
  if (do_ioctl (kbd_fd, VT_RELDISP, 1))	/* switch ok by me */
    v_printf("VID: VT_RELDISP failed!\n");
}

static void allow_switch (void)
{
  v_printf("VID: allow_switch\n");
  scr_state.vt_allow = 1;
  if (scr_state.vt_requested) {
    v_printf("VID: clearing old vt request\n");
    SIGRELEASE_call();
  }
}

static int wait_vc_active (void)
{
  if (ioctl (kbd_fd, VT_WAITACTIVE, scr_state.console_no) < 0) {
    if (errno != EINTR) {
      error ("ERROR: VT_WAITACTIVE for %d gave %d: %s\n", scr_state.console_no,
	     errno, strerror (errno));
    }
    return -1;
  }
  else
    return 0;
}

static void release_vt(int sig, struct sigcontext_struct context)
{
  v_printf("VID: release_vt\n");
  signal_queue(SIGRELEASE_call);
}

#if 0
/* allows remapping even if memory is mapped in...this is useful, as it
 * remembers where the video mem *was* mapped, unmaps from that, and then
 * remaps it to where the text page number says it should be
 */
void get_video_ram(int waitflag)
{
  char *graph_mem;
  char *sbase;
  size_t ssize;

  char *textbuf = NULL, *vgabuf = NULL;

  v_printf ("get_video_ram STARTED\n");
  if (config.vga) {
    ssize = GRAPH_SIZE;
    sbase = (char *) GRAPH_BASE;
  }
  else {
    ssize = TEXT_SIZE;
    sbase = PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE));
  }

  if (waitflag == WAIT) {
    v_printf ("VID: get_video_ram WAITING\n");
    /* XXX - wait until our console is current (mixed signal functions) */
    do {
      if (!wait_vc_active ())
	break;
      v_printf ("Keeps waiting...And\n");
    }
    while (errno == EINTR);
  }

  if (config.vga) {
    if (READ_BYTE(BIOS_VIDEO_MODE) == 3 && READ_BYTE(BIOS_CURRENT_SCREEN_PAGE) < 8) {
      textbuf = malloc (TEXT_SIZE * 8);
      memcpy (textbuf, PAGE_ADDR (0), TEXT_SIZE * 8);
    }

    if (scr_state.mapped) {
      vgabuf = malloc (GRAPH_SIZE);
      memcpy (vgabuf, (caddr_t) GRAPH_BASE, GRAPH_SIZE);
      graph_mem = (char *) mmap ((caddr_t) GRAPH_BASE,
				 GRAPH_SIZE,
				 PROT_EXEC | PROT_READ | PROT_WRITE,
				 MAP_PRIVATE | MAP_FIXED | MAP_ANON,
				 -1,
				 0);
      memcpy ((caddr_t) GRAPH_BASE, vgabuf, GRAPH_SIZE);
    }
  }
  else {
    textbuf = malloc (TEXT_SIZE);
    memcpy (textbuf, scr_state.virt_address, TEXT_SIZE);

    if (scr_state.mapped) {
      graph_mem = (char *) mmap ((caddr_t) scr_state.virt_address,
				 TEXT_SIZE,
				 PROT_EXEC | PROT_READ | PROT_WRITE,
				 MAP_PRIVATE | MAP_FIXED | MAP_ANON,
				 -1,
				 0);
      memcpy (scr_state.virt_address, textbuf, TEXT_SIZE);
    }
  }
  scr_state.mapped = 0;

  if (config.vga) {
    if (READ_BYTE(BIOS_VIDEO_MODE) == 3) {
      if (dosemu_regs.mem && textbuf)
	memcpy (dosemu_regs.mem, textbuf, dosemu_regs.save_mem_size[0]);
      /*      else error("ERROR: no dosemu_regs.mem!\n"); */
    }
    g_printf ("mapping GRAPH_BASE\n");
    graph_mem = (char *) mmap ((caddr_t) GRAPH_BASE,
			       (size_t) (GRAPH_SIZE),
			       PROT_READ | PROT_WRITE,
			       MAP_SHARED | MAP_FIXED,
			       mem_fd,
			       GRAPH_BASE);
    
    /* the code below is done by the video save/restore code */
  }
  else {
    /* this is used for page switching */
    if (PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE)) != scr_state.virt_address)
      memcpy (textbuf, PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE)), TEXT_SIZE);
    
    g_printf ("mapping PAGE_ADDR\n");
    graph_mem = (char *) mmap ((caddr_t) PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE)),
			       TEXT_SIZE,
			       PROT_READ | PROT_WRITE,
			       MAP_SHARED | MAP_FIXED,
			       mem_fd,
			       phys_text_base);
    
    if ((long) graph_mem < 0) {
      error ("ERROR: mmap error in get_video_ram (text)\n");
      return;
    }
    else
      v_printf ("CONSOLE VIDEO address: %p %p %p\n", (void *) graph_mem,
		(void *) phys_text_base, (void *) PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE)));
    
    /* copy contents of page onto video RAM */
    memcpy ((caddr_t) PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE)), textbuf, TEXT_SIZE);
  }

  if (vgabuf)
    free (vgabuf);
  if (textbuf)
    free (textbuf);
  
  scr_state.pageno = READ_BYTE(BIOS_CURRENT_SCREEN_PAGE);
  scr_state.virt_address = PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE));
  scr_state.phys_address = graph_mem;
  scr_state.mapped = 1;
}
#endif

/* this puts the VC under process control */
void set_process_control (void)
{
  struct vt_mode vt_mode;
  struct sigaction sa;

  vt_mode.mode = VT_PROCESS;
  vt_mode.waitv = 0;
  vt_mode.relsig = SIG_RELEASE;
  vt_mode.acqsig = SIG_ACQUIRE;
  vt_mode.frsig = 0;

  scr_state.vt_requested = 0;	/* a switch has not been attempted yet */
  allow_switch();

  NEWSETQSIG(SIG_RELEASE, release_vt);
  NEWSETQSIG(SIG_ACQUIRE, acquire_vt);

  if (do_ioctl(kbd_fd, VT_SETMODE, (int) &vt_mode))
    v_printf("VID: initial VT_SETMODE failed!\n");
  v_printf("VID: Set process control\n");
}

void clear_process_control (void)
{
  struct vt_mode vt_mode;

  v_printf ("VID: Clearing process control\n");

  vt_mode.mode = VT_AUTO;
  ioctl (kbd_fd, VT_SETMODE, (int) &vt_mode);
  signal (SIG_RELEASE, SIG_IGN);
  signal (SIG_ACQUIRE, SIG_IGN);
}

void set_vc_screen_page (int page)
{
  v_printf("VID: set_vc_screen_page(%d)\n", page);

  WRITE_BYTE(BIOS_CURRENT_SCREEN_PAGE, page);

  /* okay, if we have the current console, and video ram is mapped.
   * this has to be "atomic," or there is a "race condition": the
   * user may change consoles between our check and our remapping
   */
#if 0
  forbid_switch();
  if (scr_state.mapped)		/* should I check for foreground VC instead? */
    get_video_ram (WAIT);
  allow_switch();
#endif
}

void activate(int con_num)
{
  if (in_ioctl) {
    k_printf("KBD: can't ioctl for activate, in a signal handler\n");
    do_ioctl(kbd_fd, VT_ACTIVATE, con_num);
  } else
    do_ioctl(kbd_fd, VT_ACTIVATE, con_num);
}
